#  Copyright (c) 2021 Rikai Authors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Custom Mlflow model logger to make sure models have the right logging for
Rikai SQL ML
"""
import warnings
from typing import Any, Optional

CONF_MLFLOW_TRACKING_URI = "rikai.sql.ml.registry.mlflow.tracking_uri"
CONF_MLFLOW_OUTPUT_SCHEMA = "rikai.output.schema"
CONF_MLFLOW_SPEC_VERSION = "rikai.spec.version"
CONF_MLFLOW_PRE_PROCESSING = "rikai.transforms.pre"
CONF_MLFLOW_POST_PROCESSING = "rikai.transforms.post"
CONF_MLFLOW_MODEL_FLAVOR = "rikai.model.flavor"
CONF_MLFLOW_ARTIFACT_PATH = "rikai.model.artifact_path"


class MlflowLogger:
    """
    An alternative model logger for use during training instead of the vanilla

    """

    _CURRENT_MODEL_SPEC_VERSION = "1.0"

    def __init__(self, flavor: str):
        """
        Parameters
        ----------
        flavor: str
            The model flavor
        """
        self.spec_version = MlflowLogger._CURRENT_MODEL_SPEC_VERSION
        self.flavor = flavor

    def log_model(
        self,
        model: Any,
        artifact_path: str,
        schema: str,
        pre_processing: Optional[str] = None,
        post_processing: Optional[str] = None,
        registered_model_name: Optional[str] = None,
        **kwargs,
    ):
        """Convenience function to log the model with tags needed by rikai.
        This should be called during training when the model is produced.

        Parameters
        ----------
        model: Any
            The model artifact object
        artifact_path: str
            The relative (to the run) artifact path
        schema: str
            Output schema (pyspark DataType)
        pre_processing: str, default None
            Full python module path of the pre-processing transforms
        post_processing: str, default None
            Full python module path of the post-processing transforms
        registered_model_name: str, default None
            Model name in the mlflow model registry
        kwargs: dict
            Passed to `mlflow.<flavor>.log_model`

        Examples
        --------
        # import mlflow.pytorch  <= vanilla mlflow usage
        import rikai.mlflow

        # Log PyTorch model
        with mlflow.start_run() as run:

            # Training loop
            # ...

            # Assume `model` is the trained model from the training loop
            rikai.mlflow.pytorch.log_model(model, "model",
                      schema=rikai.contrib.torch.transforms.fasterrcnn_resnet50_fpn.OUTPUT_SCHEMA,
                      pre_processing="rikai.contrib.torch.transforms.fasterrcnn_resnet50_fpn.pre_processing",
                      post_processing="rikai.contrib.torch.transforms.fasterrcnn_resnet50_fpn.post_processing",
                      registered_model_name="MyPytorchModel")

        For more details see `mlflow docs <https://www.mlflow.org/docs/latest/python_api/mlflow.pytorch.html#mlflow.pytorch.log_model>`_.
        """  # noqa E501

        try:
            import mlflow
        except ImportError as e:
            raise ImportError(
                "Couldn't import mlflow. Please make sure to "
                "`pip install mlflow` explicitly or install "
                "the correct extras like `pip install rikai[mlflow]`"
            ) from e

        # no need to set the tracking uri here since this is intended to be
        # called inside the training loop within mlflow.start_run
        getattr(mlflow, self.flavor).log_model(
            model,
            artifact_path,
            registered_model_name=registered_model_name,
            **kwargs,
        )
        tags = {
            CONF_MLFLOW_SPEC_VERSION: MlflowLogger._CURRENT_MODEL_SPEC_VERSION,
            CONF_MLFLOW_MODEL_FLAVOR: self.flavor,
            CONF_MLFLOW_OUTPUT_SCHEMA: schema,
            CONF_MLFLOW_PRE_PROCESSING: pre_processing,
            CONF_MLFLOW_POST_PROCESSING: post_processing,
            CONF_MLFLOW_ARTIFACT_PATH: artifact_path,
        }
        for k in (CONF_MLFLOW_PRE_PROCESSING, CONF_MLFLOW_POST_PROCESSING):
            if not tags[k]:
                del tags[k]
                warnings.warn(
                    f"value of {k} is None or empty and "
                    "will not be populated to MLflow"
                )
        mlflow.set_tags(tags)


KNOWN_FLAVORS = ["pytorch", "sklearn"]
for flavor in KNOWN_FLAVORS:
    globals()[flavor] = MlflowLogger(flavor)



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from pathlib import Path
from typing import Any, Dict, Optional, Union
from urllib.parse import urlparse

import yaml
from pyspark.sql import SparkSession

from rikai.io import open_uri
from rikai.logging import logger
from rikai.spark.sql.codegen.base import (
    ModelSpec,
    register_udf,
    Registry,
    udf_from_spec,
)
from rikai.spark.sql.exceptions import SpecError

__all__ = ["FileSystemRegistry"]


class FileModelSpec(ModelSpec):
    """Model Spec.

    Parameters
    ----------
    spec_uri : str or Path
        Spec file URI
    options : Dict[str, Any], optional
        Additionally options. If the same option exists in spec already,
        it will be overridden.
    validate : bool, default True.
        Validate the spec during construction. Default ``True``.
    """

    def __init__(
        self,
        spec_uri: Union[str, Path],
        options: Optional[Dict[str, Any]] = None,
        validate: bool = True,
    ):
        with open_uri(spec_uri) as fobj:
            spec = yaml.load(fobj, Loader=yaml.FullLoader)
        self.base_dir = os.path.dirname(spec_uri)
        spec.setdefault("options", {})
        if options:
            spec["options"].update(options)
        super().__init__(spec, validate=validate)

    def load_model(self):
        if self.flavor == "pytorch":
            from rikai.spark.sql.codegen.pytorch import load_model_from_uri

            return load_model_from_uri(self.model_uri)
        else:
            raise SpecError("Unsupported flavor {}".format(self.flavor))

    @property
    def model_uri(self):
        """Absolute model URI."""
        origin_uri = super().model_uri
        parsed = urlparse(origin_uri)
        if parsed.scheme or os.path.isabs(origin_uri):
            return origin_uri
        return os.path.join(self.base_dir, origin_uri)


def codegen_from_yaml(
    spark: SparkSession,
    spec_uri: str,
    name: Optional[str] = None,
    options: Optional[Dict[str, str]] = None,
) -> str:
    """Generate code from a YAML file.

    Parameters
    ----------
    spark : SparkSession
        A live spark session
    spec_uri : str
        the model spec URI
    name : model name
        The name of the model.
    options : dict
        Optional parameters passed to the model.

    Returns
    -------
    str
        Spark UDF function name for the generated data.
    """
    spec = FileModelSpec(spec_uri, options=options)
    udf = udf_from_spec(spec)
    return register_udf(spark, udf, name)


class FileSystemRegistry(Registry):
    """FileSystem-based Model Registry"""

    def __init__(self, spark: SparkSession):
        self._spark = spark
        self._jvm = spark.sparkContext._jvm

    def __repr__(self):
        return "FileSystemRegistry"

    def resolve(self, spec):
        name = spec.getName()
        uri = spec.getUri()
        options = spec.getOptions()
        logger.info(f"Resolving model {name} from {uri}")
        if uri.endswith(".yml") or uri.endswith(".yaml"):
            func_name = codegen_from_yaml(self._spark, uri, name, options)
        else:
            raise ValueError(f"Model URI is not supported: {uri}")

        model = self._jvm.ai.eto.rikai.sql.model.SparkUDFModel(
            name, uri, func_name
        )
        # TODO: set options
        return model



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
This is the implementation at https://pytorch.org/hub/ultralytics_yolov5/
With autoShape it does not require pre-processing
However it is NOT testable as it is not packaged as a python repository.
To use this make sure you install yolov5 and yolov5 requirements

```
pip install -qr https://raw.githubusercontent.com/ultralytics/yolov5/master/requirements.txt
pip install yolov5
```
"""  # noqa E501

from typing import Any, Callable, Dict

__all__ = ["post_processing", "OUTPUT_SCHEMA"]


def post_processing(options: Dict[str, Any]) -> Callable:
    def post_process_func(batch: "Detections"):
        """
        Parameters
        ----------
        batch: Detections
            The ultralytics yolov5 (in torch hub) autoShape output
        """
        results = []
        for predicts in batch.pred:
            predict_result = {
                "boxes": [],
                "label_ids": [],
                "scores": [],
            }
            for *box, conf, cls in predicts.tolist():
                predict_result["boxes"].append(box)
                predict_result["label_ids"].append(cls)
                predict_result["scores"].append(conf)
            results.append(predict_result)
        return results

    return post_process_func


OUTPUT_SCHEMA = (
    "struct<boxes:array<array<float>>, scores:array<float>, "
    "label_ids:array<int>>"
)



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import secrets
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Optional

from jsonschema import validate, ValidationError
from pyspark.sql import SparkSession

from rikai.internal.reflection import find_class
from rikai.logging import logger
from rikai.spark.sql.exceptions import SpecError
from rikai.spark.sql.schema import parse_schema

__all__ = ["Registry", "ModelSpec"]


def _identity(x):
    return x


class Registry(ABC):
    """Base class of a Model Registry"""

    @abstractmethod
    def resolve(self, spec: "ModelSpec"):
        """Resolve a model from a model URI.

        Parameters
        ----------
        spec : ModelSpec
        """


# JSON schema specification for the model specifications
# used to validate model spec input
MODEL_SPEC_SCHEMA = {
    "type": "object",
    "properties": {
        "version": {
            "type": "string",
            "description": "Model SPEC format version",
        },
        "name": {"type": "string", "description": "Model name"},
        "schema": {"type": "string"},
        "model": {
            "type": "object",
            "description": "model description",
            "properties": {
                "uri": {"type": "string"},
                "flavor": {"type": "string"},
            },
            "required": ["uri"],
        },
        "transforms": {
            "type": "object",
            "properties": {
                "pre": {"type": "string"},
                "post": {"type": "string"},
            },
        },
    },
    "required": ["version", "schema", "model"],
}


class ModelSpec(ABC):
    """Model Spec.

    Parameters
    ----------
    spec : dict
        Dictionary representation of an input spec
    validate : bool, default True.
        Validate the spec during construction. Default ``True``.
    """

    def __init__(self, spec: Dict[str, Any], validate: bool = True):
        self._spec = spec
        self._spec["options"] = self._spec.get("options", {})
        if validate:
            self.validate()

    def validate(self):
        """Validate model spec

        Raises
        ------
        SpecError
            If the spec is not well-formatted.
        """
        logger.debug("Validating spec: %s", self._spec)
        try:
            validate(instance=self._spec, schema=MODEL_SPEC_SCHEMA)
        except ValidationError as e:
            raise SpecError(e.message) from e

    @property
    def version(self) -> str:
        """Returns spec version."""
        return str(self._spec["version"])

    @property
    def name(self) -> str:
        """Return model name"""
        return self._spec["name"]

    @property
    def model_uri(self) -> str:
        """Return Model artifact URI"""
        return self._spec["model"]["uri"]

    @abstractmethod
    def load_model(self) -> Any:
        """Load the model artifact specified in this spec"""

    @property
    def flavor(self) -> str:
        """Model flavor"""
        return self._spec["model"].get("flavor", "")

    @property
    def schema(self) -> str:
        """Return the output schema of the model."""
        return parse_schema(self._spec["schema"])

    @property
    def options(self) -> Dict[str, Any]:
        """Model options"""
        return self._spec["options"]

    @property
    def pre_processing(self) -> Optional[Callable]:
        """Return pre-processing transform if exists"""
        if (
            "transforms" not in self._spec
            or "pre" not in self._spec["transforms"]
            or self._spec["transforms"]["pre"] is None
        ):
            # Passthrough
            return _identity
        f = find_class(self._spec["transforms"]["pre"])
        return f(self.options)

    @property
    def post_processing(self) -> Optional[Callable]:
        """Return post-processing transform if exists"""
        if (
            "transforms" not in self._spec
            or "post" not in self._spec["transforms"]
            or self._spec["transforms"]["post"] is None
        ):
            # Passthrough
            return _identity
        f = find_class(self._spec["transforms"]["post"])
        return f(self.options)


def udf_from_spec(spec: ModelSpec):
    """Return a UDF from a given ModelSpec

    Parameters
    ----------
    spec : ModelSpec
        A model spec

    Returns
    -------
    str
        Spark UDF function name for the generated data.
    """
    if spec.version != "1.0":
        raise SpecError(
            f"Only spec version 1.0 is supported, got {spec.version}"
        )

    if spec.flavor == "pytorch":
        from rikai.spark.sql.codegen.pytorch import generate_udf

        return generate_udf(spec)
    elif spec.flavor == "sklearn":
        from rikai.spark.sql.codegen.sklearn import generate_udf

        return generate_udf(spec)
    else:
        raise SpecError(f"Unsupported model flavor: {spec.flavor}")


def register_udf(spark: SparkSession, udf: Callable, name: str) -> str:
    """
    Register a given UDF with the give Spark session under the given name.
    """
    func_name = f"{name}_{secrets.token_hex(4)}"
    spark.udf.register(func_name, udf)
    logger.info(f"Created model inference pandas_udf with name {func_name}")
    return func_name



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Callable, Dict

from torchvision import transforms as T

from rikai.contrib.torch.transforms.utils import uri_to_pil

__all__ = ["pre_processing", "post_processing"]

DEFAULT_MIN_SCORE = 0.5


def pre_processing(options: Dict[str, Any]) -> Callable:
    return T.Compose(
        [
            uri_to_pil,
            T.ToTensor(),
        ]
    )


def post_processing(options: Dict[str, Any]) -> Callable:
    min_score = float(options.get("min_score", DEFAULT_MIN_SCORE))

    def post_process_func(batch):
        results = []
        for predicts in batch:
            predict_result = {
                "boxes": [],
                "label_ids": [],
                "scores": [],
            }
            for box, label, score in zip(
                predicts["boxes"].tolist(),
                predicts["labels"].tolist(),
                predicts["scores"].tolist(),
            ):
                if score < min_score:
                    continue
                predict_result["boxes"].append(box)
                predict_result["label_ids"].append(label)
                predict_result["scores"].append(score)

            results.append(predict_result)
        return results

    return post_process_func


OUTPUT_SCHEMA = (
    "struct<boxes:array<array<float>>, scores:array<float>, "
    "label_ids:array<int>>"
)
